"""
Predefined couplings of interest for most applications.

"""

import jax.numpy as np


def make_delayed_coupling(weights, delay_steps, pre, post, nh, isvar):
    """
    Construct a dense delayed coupling function. 

    Parameters
    ==========
    weights : array
        Coupling weights.
    delay_steps : array
        Number of delay steps per connection i, j.

    ...
    To be finished
    ...

    Notes
    =====

    - This construction assumes a particular layout for the history
      buffer: xt.shape == (nh+1+nt, nsvar, nnode, ...). 

    """
    nn = weights.shape[0]
    nodes = np.tile(np.r_[:nn], (nn, 1))
    def cfun(t, xt, x, params):
        dx = xt[nh + t - delay_steps, isvar, nodes]
        xij = pre(dx, x, params)
        gx = (weights * xij).sum(axis=1)
        return post(gx, params)
    return cfun


def make_linear_cfun(SC, a=1.0, b=0.0):
    """Construct a linear coupling function with slope `a` and offset `b`.
    """
    def cfun(xj):
        "Compute linear coupling for pre-synaptic state variables xj."
        if xj.ndim == 1:  # no delays
            gx = SC @ xj
        elif xj.ndim == 2:  # delays
            gx = np.sum(SC * xj, axis=1)
        return a*gx + b
    return cfun


def make_diff_cfun(SC, a=1.0, b=0.0):
    """Construct a linear difference coupling."""
    nn = np.r_[:SC.shape[0]]
    diffdiag = np.diag(SC) - SC.sum(axis=1)
    SC_ = SC.at[nn,nn].set(diffdiag)
    # fix diagonal according to trick
    return make_linear_cfun(SC_, a=a, b=b)



"""
Functions for building time stepping loops.

"""

import jax
import jax.numpy as np


def heun_step(x, dfun, dt, *args, add=0):
    """Use a Heun scheme to step state with a right hand sides dfun(.)
    and additional forcing term add.
    """
    d1 = dfun(x, *args)
    d2 = dfun(x + dt*d1 + add, *args)
    return x + dt*0.5*(d1 + d2) + add

def make_sde(dt, dfun, gfun):
    """Use a stochastic Heun scheme to integrate autonomous stochastic
    differential equations (SDEs).

    Parameters
    ==========
    dt : float
        Time step
    dfun : function
        Function of the form `dfun(x, p)` that computes drift coefficients of
        the stochastic differential equation.
    gfun : function or float
        Function of the form `gfun(x, p)` that computes diffusion coefficients
        of the stochastic differential equation. If a numerical value is
        provided, this is used as a constant diffusion coefficient for additive
        linear SDE.

    Returns
    =======
    step : function
        Function of the form `step(x, z_t, p)` that takes one step in time
        according to the Heun scheme.
    loop : function
        Function of the form `loop(x0, zs, p)` that iteratively calls `step`
        for all `z`.

    Notes
    =====

    In both cases, a Jax compatible parameter set `p` is provided, either an array
    or some pytree compatible structure.

    Note that the integrator does not sample normally distributed noise, so this
    must be provided by the user.


    >>> import vbjax as vb
    >>> _, sde = vb.make_sde(1.0, lambda x, p: -x, 0.1)
    >>> sde(1.0, vb.randn(4), None)
    Array([ 0.5093468 ,  0.30794007,  0.07600437, -0.03876263], dtype=float32)

    """

    sqrt_dt = np.sqrt(dt)

    # gfun is a numerical value or a function f(x,p) -> sig
    if not hasattr(gfun, '__call__'):
        sig = gfun
        gfun = lambda *_: sig

    def step(x, z_t, p):
        noise = gfun(x, p) * sqrt_dt * z_t
        return heun_step(x, dfun, dt, p, add=noise)

    @jax.jit
    def loop(x0, zs, p):
        def op(x, z):
            x = step(x, z, p)
            return x, x
        return jax.lax.scan(op, x0, zs)[1]

    return step, loop


def make_ode(dt, dfun):
    """Use a Heun scheme to integrate autonomous ordinary differential
    equations (ODEs).

    Parameters
    ==========
    dt : float
        Time step
    dfun : function
        Function of the form `dfun(x, p)` that computes derivatives of the
        ordinary differential equations.

    Returns
    =======
    step : function
        Function of the form `step(x, t, p)` that takes one step in time
        according to the Heun scheme.
    loop : function
        Function of the form `loop(x0, ts, p)` that iteratively calls `step`
        for all time steps `ts`.

    Notes
    =====

    In both cases, a Jax compatible parameter set `p` is provided, either an array
    or some pytree compatible structure.

    >>> import vbjax as vb, jax.numpy as np
    >>> _, ode = vb.make_ode(1.0, lambda x, p: -x)
    >>> ode(1.0, np.r_[:4], None)
    Array([0.5   , 0.25  , 0.125 , 0.0625], dtype=float32, weak_type=True)

    """

    def step(x, t, p):
        return heun_step(x, dfun, dt, p)

    @jax.jit
    def loop(x0, ts, p):
        def op(x, t):
            x = step(x, t, p)
            return x, x
        return jax.lax.scan(op, x0, ts)[1]

    return step, loop


def make_dde(dt, nh, dfun, unroll=10):
    "Invokes make_sdde w/ gfun 0."
    return make_sdde(dt, nh, dfun, 0, unroll)


def make_sdde(dt, nh, dfun, gfun, unroll=1, zero_delays=False):
    """Use a stochastic Heun scheme to integrate autonomous
    stochastic delay differential equations (SDEs).

    Parameters
    ==========
    dt : float
        Time step
    nh : int
        Maximum delay in time steps.
    dfun : function
        Function of the form `dfun(xt, x, t, p)` that computes drift coefficients of
        the stochastic differential equation.
    gfun : function or float
        Function of the form `gfun(x, p)` that computes diffusion coefficients
        of the stochastic differential equation. If a numerical value is
        provided, this is used as a constant diffusion coefficient for additive
        linear SDE.

    Returns
    =======
    step : function
        Function of the form `step((x_t,t), z_t, p)` that takes one step in time
        according to the Heun scheme.
    loop : function
        Function of the form `loop((xs, t), p)` that iteratively calls `step`
        for each `xs[nh:]` and starting time index `t`.

    Notes
    =====

    - A Jax compatible parameter set `p` is provided, either an array
      or some pytree compatible structure.
    - The integrator does not sample normally distributed noise, so this
      must be provided by the user.
    - The history buffer passed to the user functions, on the corrector
      stage of the Heun method, does not contain the predictor stage, for
      performance reasons, unless `zero_delays` is set to `True`.
      A good compromise can be to set all zero delays to dt.

    >>> import vbjax as vb, jax.numpy as np
    >>> _, sdde = vb.make_sdde(1.0, 2, lambda xt, x, t, p: -xt[t-2], 0.0)
    >>> x,t = sdde(np.ones(6)+10, None)
    >>> x
    Array([ 11.,  11.,  11.,   0., -11., -22.], dtype=float32)

    """

    heun = True
    sqrt_dt = np.sqrt(dt)

    # TODO nh == 0: return make_sde(dt, dfun, gfun) etc

    # gfun is a numerical value or a function f(x,p) -> sig
    if not hasattr(gfun, '__call__'):
        sig = gfun
        gfun = lambda *_: sig

    def step(xt_t, z_t, p):
        xt, t = xt_t
        x = xt[nh + t]
        noise = gfun(x, p) * sqrt_dt * z_t
        d1 = dfun(xt, x, t, p)
        xi = x + dt*d1 + noise
        if heun:
            if zero_delays:
                # severe performance hit (5x+)
                xt = xt.at[nh + t + 1].set(xi)
            d2 = dfun(xt, xi, t+1, p)
            nx = x + dt*0.5*(d1 + d2) + noise
        else:
            nx = xi
        xt = xt.at[nh + t + 1].set(nx)
        return (xt,t+1), nx

    @jax.jit
    def loop(xt, p, t=0):
        "xt is the buffer, zt is (ts, zs), p is parameters."
        op = lambda xt, tz: step(xt, tz, p)
        return jax.lax.scan(op, (xt,t), xt[nh:], unroll=unroll)[0]

    return step, loop




import jax
import jax.numpy as np
from .loops import heun_step
from .neural_mass import BOLDTheta, bold_dfun


def make_offline(step_fn, sample_fn, *args):
    "Compute monitor samples in an offline or batch fashion."
    def op(mon, x):
        mon = step_fn(mon, x)
        return mon, None
    def offline_sample(mon, xs):
        mon, _ = jax.lax.scan(op, mon, xs)
        mon, samp = sample_fn(mon)
        return mon, samp
    return offline_sample

# NB shape here is the input shape of neural activity

def make_timeavg(shape):
    "Make a time average monitor."
    new = lambda : {'y': np.zeros(shape), 'n': 0}
    def step(buf, x):
        return {'y': buf['y'] + x,
                'n': buf['n'] + 1}
    def sample(buf):
        return new(), buf['y'] / buf['n']
    return new(), step, sample


def compute_sarvas_gain(q, r, o, att, Ds=0, Dc=0) -> np.ndarray:
    # https://gist.github.com/maedoc/add7c3206f81d59105753a04f7c1fcf4
    pass


def make_gain(gain, shape=None):
    "Make a gain-matrix monitor suitable for sEEG, EEG & MEG."
    tavg_shape = gain.shape[:1] + (shape[1:] if shape else ())
    buf, tavg_step, tavg_sample = make_timeavg(tavg_shape)
    step = lambda b, x: tavg_step(b, gain @ x)
    return buf, step, tavg_sample


def make_bold(shape, dt, p: BOLDTheta):
    "Make a BOLD fMRI monitor."
    sfvq = np.ones((4,) + shape)
    sfvq = sfvq.at[0].set(0)
    def step(sfvq, x):
        return heun_step(sfvq, bold_dfun, dt, x, p)
    def sample(buf):
        s, f, v, q = buf
        return buf, p.v0 * (p.k1*(1 - q) + p.k2*(1 - q / v) + p.k3*(1 - v))
    return sfvq, step, sample


def make_cov(shape):
    # welford online cov estimate yields o(1) backprop memory usage
    # https://github.com/maedoc/tvb-fut/blob/master/lib/github.com/maedoc/tvb-fut/stats.fut#L9
    assert len(shape) == 1, "FC over 1-D vectors for now."
    cov_shape = shape * 2
    def new():
        buf = {'count': 0,
               'mean': np.zeros(shape),
               'cov': np.zeros(cov_shape)
               }
        return buf
    def step(buf, x):
        count, mean, cov = buf['count'], buf['mean'], buf['cov']
        count = count + 1
        dx = x - mean
        mean = mean + dx / count
        cov = cov + np.outer(dx, dx)
        # count = count + 1
        # delta = val - mean
        # mean = mean + delta / count
        # delta2 = val - mean
        # m2 = m2 + delta * delta2
        return {'count': count,
                'mean': mean,
                'cov': cov}
    def sample(buf):
        var = buf['cov'] / buf['count']
        # std = np.sqrt(var)
        return new(), var
    return new(), step, sample


def make_fc(shape):
    new_buf, step, cov_sample = make_cov(shape)
    def fc_sample(buf):
        buf, cov = cov_sample(buf)
        # Rij = Cij / sqrt(Cii Cjj)
        cii = np.diag(cov)
        return new_buf.copy(), cov / np.sqrt(np.outer(cii, cii))
    return new_buf.copy(), step, fc_sample


def make_fft(shape, period):
    pass

# TODO sliding window versions of those

# @jax.checkpoint to lower memory usage?



import numpy
import jax
import jax.dlpack


def to_np(x: jax.numpy.ndarray) -> numpy.ndarray:
    if jax.default_backend() == 'gpu':
        return numpy.array(x)
    return numpy.from_dlpack(x)

def to_jax(x: numpy.ndarray):
    "Move NumPy array to JAX via DLPack."
    x_dlp = x.__dlpack__()
    x_jax: jax.numpy.ndarray = jax.dlpack.from_dlpack(x_dlp)
    # if jax.default_backend() == 'gpu':
    #     x_jax = jax.device_put(x_jax)
    return x_jax


print('vbjax ███▒▒▒▒▒▒▒ loading')

# default to setting up many cores
def _use_many_cores():
    import os, sys
    import multiprocessing as mp
    if 'XLA_FLAGS' not in os.environ:
        n = mp.cpu_count()
        value = '--xla_force_host_platform_device_count=%d' % n
        os.environ['XLA_FLAGS'] = value
        print(f'vbjax (ﾉ☉ヮ⚆)ﾉ ⌒*:･ﾟ✧ can haz {n} cores')
    else:
        print('vbjax XLA_FLAGS already set\n')

_use_many_cores()

# import stuff
from .loops import make_sde, make_ode, make_dde, make_sdde, heun_step
from .shtlc import make_shtdiff
from .neural_mass import (
        JRState, JRTheta, jr_dfun, jr_default_theta,
        MPRState, MPRTheta, mpr_dfun, mpr_default_theta,
        BOLDTheta, compute_bold_theta, bold_default_theta, bold_dfun
        )
from .regmap import make_region_mapping
from .coupling import (
        make_diff_cfun, make_linear_cfun, make_delayed_coupling
        )
from .connectome import make_conn_latent_mvnorm
from .sparse import make_spmv, csr_to_jax_bcoo, make_sg_spmv
from .monitor import (
    make_timeavg, make_bold, make_gain, make_offline, make_cov, make_fc)
from .layers import make_dense_layers
from .diagnostics import shrinkage_zscore
from .embed import embed_neural_flow, embed_polynomial, embed_gradient, embed_autoregress
from .util import to_jax, to_np
from ._version import __version__

# some random setup for convenience
from jax import random
from jax import numpy as np
key = random.PRNGKey(42)
keys = random.split(key, 100)
def randn(*shape, key=key):
    return random.normal(key, shape)

def randn_connectome(n, key=key):
    w, l = np.abs(randn(2, n, n, key=key))
    l = l * 100
    return w+w.T, l+l.T

# some simple plots
def plot_states(xs, names=None, jpg=None, show=False):
    import pylab as pl
    names = names or ['x%d' % i for i in range(xs.shape[1])]
    for i in range(xs.shape[1]):
        pl.subplot(xs.shape[1], 1, i+1)
        pl.plot(xs[:, i], 'k', alpha=0.3)
        pl.ylabel(names[i])
        pl.xlabel('time')
        pl.grid(1)
    pl.tight_layout()
    if jpg:
        pl.savefig(jpg if '.jpg' in jpg else jpg + '.jpg')
    if show:
        pl.show()

def make_field_gif(xt, gifname, fps=15):
    import pylab as pl
    import matplotlib.animation as animation
    fig, ax = pl.subplots()
    fig.set_size_inches((4,2))
    im = ax.imshow(xt[0,0])
    ax.set_xlabel('Longitude')
    ax.set_ylabel('Latitude')
    ax.set_title('Field activity')
    fig.tight_layout()
    def animate(i):
        im.set_data(xt[i+1,0])
        im.set_clim()
        return im,
    ani = animation.FuncAnimation(fig, animate,
            repeat=True, frames=len(xt)-1, interval=20)
    writer = animation.PillowWriter(fps=fps, bitrate=400)
    ani.save(gifname, writer=writer)

print('vbjax ᕕ(ᐛ)ᕗ ready')


import time
import numpy as np
import jax
import jax.test_util
import jax.dlpack
import jax.numpy as jnp
import scipy.sparse
import vbjax
import pytest


def _test_spmv(spmv, A, n):
    nx = np.random.randn(n)
    jx = jnp.array(nx)

    # test the function itself
    jb = spmv(jx)
    nb = A @ nx
    np.testing.assert_allclose(jb, nb, 1e-6, 1e-6)

    # now its gradient
    jax.test_util.check_grads(spmv, (jx,), order=1, modes=('rev',))


def test_csr_scipy():
    n = 10
    A = scipy.sparse.random(n, n, density=0.1).tocsr()
    spmv = vbjax.make_spmv(A)
    _test_spmv(spmv, A, n)


def test_csr_scipy_symm():
    n = 10
    A = scipy.sparse.random(n, n, density=0.1).tocsr()
    A += A.T
    spmv = vbjax.make_spmv(A, is_symmetric=True)
    _test_spmv(spmv, A, n)


def test_sg_spmv():
    n = 100
    A = scipy.sparse.random(n, n, density=0.1).tocsr()
    spmv = vbjax.make_sg_spmv(A)
    _test_spmv(spmv, A, n)


def bench_csr_to_bcoo():
    n = 1000
    A = scipy.sparse.random(n, n, density=0.1).tocsr()
    spmv = vbjax.make_spmv(A)
    jx = jax.numpy.r_[:n].astype('f')
    jb1 = spmv(jx)

    # now convert to bcoo
    jA = vbjax.csr_to_jax_bcoo(A)
    jspmv = jax.jit(lambda x: jA @ x)
    jb2 = jspmv(jx)

    np.testing.assert_allclose(jb1, jb2, 1e-6, 1e-6)

    for f, name in zip((spmv, jspmv), 'csr bcoo'.split(' ')):
        t0 = time.time()
        for _ in range(100):
            jb2 = f(jx)
        t1 = time.time()
        print(f'{name}: {t1 - t0:.3f} s')


def create_sample_data(n=1000, density_pct=10):
    A = scipy.sparse.random(n, n, density=density_pct/100).tocsr()
    jx = jax.numpy.r_[:n].astype('f')
    return A, jx


# some performance testing values
_perf_args = 'n,density_pct,grad,impl,jit'
_perf_values = [(1000, 10), (10_000, 0.02)]

# we want to test each of the above values with grad on and off
_perf_values = [vals + (flag, impl, jit)
                for flag in (True, False)
                for impl in 'scipy jaxbcoo'.split(' ')
                for jit in (True, False)
                for vals in _perf_values]

@pytest.mark.parametrize(_perf_args, _perf_values)
def test_perf_jbcoo(benchmark, n, density_pct, grad, impl, jit):
    A, x = create_sample_data(n=n, density_pct=density_pct)

    if impl == 'scipy': # TODO enum
        spmv1 = vbjax.make_spmv(A)
    elif impl == 'jaxbcoo':
        jA = vbjax.csr_to_jax_bcoo(A)
        spmv1 = lambda x: jA @ x
    else:
        raise ValueError(impl)
    assert callable(spmv1)
    
    if grad:
        spmv2 = jax.grad(lambda x: jnp.sum(spmv1(x)))
    else:
        spmv2 = spmv1
    assert callable(spmv2)

    if jit and impl not in ('scipy', ):
        spmv3 = jax.jit(spmv2)
        spmv3(x)
    else:
        spmv3 = spmv2
    assert callable(spmv3)

    benchmark(lambda : spmv3(x))    



import numpy as np
import scipy.sparse
import jax
import jax.experimental.sparse as jsp
import vbjax as vb


def make_spmv(A, is_symmetric=False, use_scipy=False):
    """
    Make a closure for a general sparse matrix-vector multiplication.

    Parameters
    ----------
    A : scipy.sparse.csr_matrix
        Constant sparse matrix.
    is_symmetric : bool, optional, default False
        Whether matrix is symmetric.
    use_scipy: bool, optional, default False
        Use scipy.

    Returns
    -------
    spmv : function
        Function implementing spase matrix vector multiply with
        support for gradients in Jax.

    """
    AT = A.T.copy()
    @jax.custom_vjp
    def matvec(x):          return vb.to_jax(A @ vb.to_np(x))
    def matvec_tr(x):       return vb.to_jax(AT @ vb.to_np(x))
    def matvec_fwd(x):      return matvec(x), None
    def matvec_bwd(res, g): return matvec(g) if is_symmetric else matvec_tr(g),
    matvec.defvjp(matvec_fwd, matvec_bwd)
    return matvec


def csr_to_jax_bcoo(A: scipy.sparse.csr_matrix) :
    "Convert CSR format to batched COO format."
    # first convert CSR to COO
    coo = A.tocoo()
    # now convert to batched COO
    data = jax.numpy.array(coo.data)
    indices = jax.numpy.array(np.c_[coo.row, coo.col])
    shape = A.shape
    return jsp.BCOO((data, indices), shape=shape)


def make_sg_spmv(A: scipy.sparse.csr_matrix):
    "Make a SpMV kernel w/ generic scatter-gather operations."
    import jax.numpy as np
    nrow = A.shape[0]
    col = np.array(A.indices)
    row = np.zeros_like(col)
    for i, (r0, r1) in enumerate(zip(A.indptr[:-1], A.indptr[1:])):
        row = row.at[r0:r1].set(i)
    dat = np.array(A.data)
    def spmv(x):
        return np.zeros(nrow).at[row].add(dat * x[col])
    return spmv


"""
Tests and benchmarks for a more generic uber loop.

- constant args
- time-dep args
- jit-time constants vs args
- monitors
- jax.checkpoint

"""

import jax
import jax.numpy as np
import vbjax as vb


def _heun_step(x, dt, f, args):
    f1 = f(x, args)
    f2 = f(x + dt*f1, args)
    return x + dt*0.5*(f1 + f2)

def _dfun1(xy, args):
    SC, a, k, stim = args
    x, y = xy
    c = np.dot(SC, x)*k
    dx = 5.0*(x - x*x*x/3 + y)
    dy = 0.2*(a - x + stim + c)
    return np.array([dx, dy])


def make_loop(dt, dfun, constants=()):
    def loop(initial_state, parameters=(), t_parameters=()):
        def step(state, t_parameters):
            args = constants + parameters + t_parameters
            next_state = _heun_step(state, dt, dfun, args)
            return next_state, next_state
        _, states = jax.lax.scan(step, initial_state, t_parameters)
        return states
    return loop


# can we go for the fully generic graph idea?
# stim(t) -> node -> monitor?
def make_loop_graph(steps, constants):
    def loop(initials, params, Tdeps):
        def body(states, tdeps):
            nexts = tuple(
                s(*args)
                for s, *args in zip(steps, constants, params, states, tdeps)
            )
            return nexts, nexts
        _, sol = jax.lax.scan(body, initials, Tdeps)
        return sol
    return loop
# would be more flexible if we had a dict or struct
# e.g. jax dataclasses




















import jax
import jax.numpy as np
import vbjax as vb


def test_sde():
    f = lambda x,_: -x
    g = lambda *_: 1e-2
    dt = 0.1
    _, run = vb.make_sde(dt, f, g)
    zs = vb.randn(100, 32)
    xs = run(zs[0]+1, zs, None)
    assert xs.shape == zs.shape


def test_ode():
    f = lambda x, _: -x
    dt = 0.1
    _, run = vb.make_ode(dt, f)
    x0 = np.r_[:32].astype('f')
    xs = run(x0, np.r_[:64], None)
    assert xs.shape == (64, 32)


def test_dde():
    def dfun(xt, x, t, p):
        xd = xt[0, t-100]
        dx = x - x**3/3 + p*xt[0, t-100]
        return dx

    _, loop = vb.make_dde(0.1, 100, dfun)
    xt0 = np.ones((1, 200))
    xt1,t = loop(xt0, 0.2)
    assert xt1.shape == (1, 200)


def test_sdde():
    import vbjax as vb, jax.numpy as np
    def dfun(xt, x, t, p):
        return -xt[t - 5]
    _, sdde = vb.make_sdde(1.0, 5, dfun, 0.01)
    sdde(vb.randn(20)+10, None)


# TODO theta method? https://gist.github.com/maedoc/c47acb9d346e31017e05324ffc4582c1


"""
These are tests of our understanding of how Jax works.  If these break
then probably things elsewhere will also break.

"""

import numpy
import jax
import jax.dlpack
import jax.test_util
import jax.numpy as np
import vbjax as vb


def test_dlpack_numpy():
    x = numpy.random.randn(10)
    jx = jax.dlpack.from_dlpack(x.__dlpack__())
    nx = numpy.from_dlpack(jx)
    numpy.testing.assert_allclose(x, nx)

def test_custom_vjp_simple():

    @jax.custom_vjp
    def foo(a, b):
        return 2*a+b, 3*a+0.5*b

    def foo_fwd(a, b):
        return foo(a, b), None

    def foo_bwd(res, g):
        g_1, g_2 = g
        return (2*g[0]+3*g[1], g[0] + 0.5*g[1])

    foo.defvjp(foo_fwd, foo_bwd)

    def bar(a, b):
        c, d = foo(a, b)
        return 2*c + 3*d

    jax.grad(bar, [0,1])(3.,4.)

    jax.test_util.check_grads(bar, (3.0, 4.0), order=1, modes=('rev',))


def test_batched_jax_norm():
    "test how to batch norms of derivatives and Jacobians"

    # this is for neural ode, so make some layers
    nn_p, nn_f = vb.make_dense_layers(3, [13])
    assert callable(nn_f)

    x = vb.randn(100, 3, 50)
    x00 = x[0,:,0]

    f = lambda x: nn_f(nn_p, x)
    assert f(x00).shape == (3, 13)  # calling w/ vector doesn't work correctly!
    assert f(x00.reshape(-1, 1))[:,0].shape == (3, )

    # let's make the shape part of the f
    f2 = lambda x: nn_f(nn_p, x.reshape(-1, 1)).reshape(-1)
    assert f2(x00).shape == (3,)
    # yess

    # this makes the Jacobian behave like we expect
    J = jax.jacfwd(f2)
    assert J(x00).shape == (3, 3)

    # then the norms should work as expected
    nf = lambda x: jax.numpy.linalg.norm(f2(x))
    nJ = lambda x: jax.numpy.linalg.norm(J(x))
    assert nf(x00).shape == nJ(x00).shape == ()

    # now vmap that over 50 time points
    bnf = jax.vmap(nf, 1)
    assert bnf(x[0,:]).shape == (50,)

    bJf = jax.vmap(nJ, 1)
    assert bJf(x[0,:]).shape == (50,)

    # and a second over 100 batch elements
    bbnf = jax.vmap(bnf, 0)
    bbJf = jax.vmap(bJf, 0)
    assert bbnf(x).shape == bbJf(x).shape == (100, 50)


def test_loop_dict():
    import jax, jax.numpy as np
    ns = {
        "int": 3,
        "float": 3.14,
        "array": np.r_[:10.0]
    }
    def op(ns, inputs):
        ns["int"] += inputs
        return ns, ns["int"]
    @jax.jit
    def loop(ns: dict, inps: np.ndarray):
        ns, _ = jax.lax.scan(op, ns, inps)
        return ns

    ns1 = loop(ns, np.r_[:3])
    ns2 = loop(ns, np.r_[2:4])
    assert ns["int"] == 3
    assert ns1["int"] == 6
    assert ns2["int"] == 8


def _dict_step(sim, t):
    x = sim['x']
    z = sim['noise'] * jax.random.normal(sim['rng'], shape=x.shape)
    f = lambda x: x - x**3/3 + sim['coupling'] * sim['weights']@x
    x = vb.heun_step(x, f, sim['dt'])
    sim['x'] = x
    eeg = sim['eeg_gain'] @ x
    return sim, eeg

def test_loop_dict2():

    # a more complicated example
    batch_size = 32
    sim = {
        'noise': 0.1,
        'coupling': 0.01,
        'dt': 0.1,
        'x': vb.randn(164, batch_size),
        'weights': vb.randn(164,164),
        'eeg_gain': vb.randn(64, 164),
        'rng': jax.random.PRNGKey(42)
    }

    def loop(sim, ts):
        sim, eegs = jax.lax.scan(_dict_step, sim, ts)
        return eegs

    params = {
        'dt': 0.01,
        'coupling': 0.01
    }
    ts = np.r_[:100]

    def loss(params, sim, ts):
        sim = sim.copy()
        sim.update(params)
        eegs = loop(sim, ts)
        return np.sum(np.square(eegs))

    jloss = jax.jit(jax.value_and_grad(loss))

    v, grads = jloss(params, sim, ts)
    for key, g_key in grads.items():
        assert np.abs(g_key) > 0


def test_loop_dict_vmap():

    sim = {
        'noise': 0.1,
        'coupling': 0.01,
        'dt': 0.1,
        'x': vb.randn(164),
        'weights': vb.randn(164,164),
        'eeg_gain': vb.randn(64, 164),
        'rng': jax.random.PRNGKey(42)
    }

    def loop(sim, ts):
        sim, eegs = jax.lax.scan(_dict_step, sim, ts)
        return eegs

    # to vmap we need specific args to vmap over
    # in simple case this is an array like initial conditions,
    def loop_x0(x0, sim, ts):
        sim = sim.copy()
        sim['x'] = x0
        return loop(sim, ts)

    ts = np.r_[:100]

    l1 = jax.jit(jax.vmap(lambda x: loop_x0(x, sim, ts)))
    l2 = jax.jit(lambda x: loop_x0(x, sim, ts))
    l3 = jax.jit(jax.vmap(lambda x: loop_x0(x, sim, ts), in_axes=1, out_axes=-1))

    a1 = vb.randn(32, 164)
    a2 = vb.randn(164, 32)

    assert l1(a1).shape == (32, ts.size, 64)
    assert l2(a2).shape == (ts.size, 64, 32)
    assert l3(a2).shape == (ts.size, 64, 32)


import numpy
import jax
import jax.numpy as np
import vbjax as vb
import pytest

def test_timeavg():
    buf, ta_step, ta_sample = vb.make_timeavg((4,))

    buf = ta_step(buf, 1.0)
    buf = ta_step(buf, np.r_[:4])
    buf, ta = ta_sample(buf)
    numpy.testing.assert_allclose(ta, (1+np.r_[:4])/2)

    buf = ta_step(buf, 2.0)
    buf, ta = ta_sample(buf)
    numpy.testing.assert_allclose(ta, 2.0)


def test_offline():
    buf, ta_step, ta_sample = vb.make_timeavg((1,))
    bufo, _, _ = vb.make_timeavg((1,))
    ta_off = vb.make_offline(ta_step, ta_sample)
    xs = vb.randn(20, 100)
    for j, x_j in enumerate(xs):
        for i, x_ji in enumerate(x_j):
            buf = ta_step(buf, x_ji)
        buf, ta = ta_sample(buf)
        bufo, tao = ta_off(bufo, x_j)
        numpy.testing.assert_allclose(tao, ta)

def test_gain():
    gain = vb.randn(4, 4)
    ones = np.ones((4, ))
    buf, g_step, g_sample = vb.make_gain(gain, ones.shape)

    buf = g_step(buf, ones)
    buf = g_step(buf, ones)
    buf = g_step(buf, ones)

    buf, eeg = g_sample(buf)

    numpy.testing.assert_allclose(eeg, gain.sum(axis=1))


def test_cov_fc():
    xs = vb.randn(1000, 16)*2 + 1.0
    fc, fc_step, fc_sample = vb.make_fc(xs[0].shape)
    fcol = vb.make_offline(fc_step, fc_sample)
    _, s_fc = fcol(fc, xs)
    numpy.testing.assert_allclose(s_fc, np.identity(16), 0.09, 0.09)


def test_bold():
    n = 8
    dt = 0.1
    p = vb.bold_default_theta
    buf, b_step, b_sample = vb.make_bold((n, ), dt, p)

    buf = b_step(buf, vb.randn(n))
    buf = b_step(buf, vb.randn(n))
    buf = b_step(buf, vb.randn(n))

    _, fmri = b_sample(buf)
    assert fmri.shape == (n,)


def setup_multiple_periods(unroll, checkpoint):

    # setup monitors
    eeg_gain = vb.randn(64, 32)
    eeg_buf, eeg_step, eeg_sample = vb.make_gain(eeg_gain)
    eeg2_buf, eeg2_step, eeg2_sample = vb.make_gain(eeg_gain)
    eeg2_offline_sample = vb.make_offline(eeg2_step, eeg2_sample)
    bold_buf, bold_step, bold_sample = vb.make_bold((eeg_gain.shape[1], ),
                                                0.1, vb.bold_default_theta)

    # our simulation state
    # TODO may be easier with jax_dataclasses
    sim = {
        'eeg_buf': eeg_buf,
        'eeg2_buf': eeg2_buf,
        'bold_buf': bold_buf,
        'freq': 0.1,
    }

    # inner scan steps neural dynamics & monitor states
    @(jax.checkpoint if checkpoint else lambda f: f)
    def op1(sim, t):
        key_t = jax.random.PRNGKey(t)
        # insert neural dynamics here
        x = jax.random.normal(key_t, shape=(eeg_gain.shape[1],))
        # update monitors
        sim['eeg_buf'] = eeg_step(sim['eeg_buf'], np.sin(x * sim['freq']))
        sim['bold_buf'] = bold_step(sim['bold_buf'], np.sin(x) * 0.25 + 1.0)
        return sim, x

    # next scan samples eeg monitors
    def op2(sim, t_):
        # sample eeg w/ period of 10*dt
        sim, raw = jax.lax.scan(op1, sim, t_ * 10 + np.r_[:10],
                                unroll=10)
        sim['eeg_buf'], eeg_t = eeg_sample(sim['eeg_buf'])
        sim['eeg2_buf'], eeg2_t = eeg2_offline_sample(sim['eeg2_buf'],
                                                      np.sin(raw * sim['freq']))
        return sim, (raw, eeg_t, eeg2_t)

    # outer scan steps from one bold sample to the next
    def op3(sim, T):
        # run for 5 samples of eeg
        sim, (raw, eeg, eeg2) = jax.lax.scan(op2, sim, T*50 + np.r_[:5],
                                       unroll=5 if unroll else 1)
        # sample fmri w/ period of 5*10*dt
        _, fmri = bold_sample(sim['bold_buf'])
        return sim, (raw, eeg, eeg2, fmri)

    return sim, op3

def test_multiple_periods():
    sim, op3 = setup_multiple_periods(False, False,)
    ts = np.r_[:10]
    sim, (raw, eeg, eeg2, fmri) = jax.lax.scan(op3, sim, ts)
    assert raw.shape == (ts.size, 5, 10, 32)
    assert eeg.shape == (ts.size, 5, 64)
    assert eeg2.shape == (ts.size, 5, 64)
    assert fmri.shape == (ts.size, 32)
    numpy.testing.assert_allclose(eeg2, eeg, 2e-3, 1e-4)


@pytest.mark.parametrize('opts', [
    f'{args} {dev}'
    for args in ['jit grad','jit grad ckp','jit','']
    for dev in 'cpu,gpu'.split(',')])
def test_multiple_periods_perf(benchmark, opts):
    opts = opts.split(' ')
    device_name = 'cpu' if 'cpu' in opts else 'gpu'
    try:
        device = jax.devices(device_name)[0]
    except RuntimeError: # no gpu
        return
    unroll = 'unroll' in opts
    with jax.default_device(device):
        sim, op3 = setup_multiple_periods(unroll, 'ckp' in opts)
        ts = np.r_[:100]
        def run(freq, sim):
            sim = sim.copy()
            sim['freq'] = freq
            sim, (raw, eeg, eeg2, fmri) = jax.lax.scan(
                op3, sim, ts,
                unroll=10 if unroll else 1)
            return np.sum(np.square(eeg))
        if 'grad' in opts:
            run = jax.grad(run)
            assert np.abs(run(0.2,sim)) > 0
        if 'jit' in opts:
            run = jax.jit(run)
            run(0.2, sim)
        benchmark(lambda : run(0.2, sim))
